from __future__ import annotations

import os
from pathlib import Path
from typing import AnyStr

import pytest

from PIL import FontFile, Image, ImageDraw, ImageFont, PcfFontFile

from .helper import (
    assert_image_equal_tofile,
    assert_image_similar_tofile,
    skip_unless_feature,
)

fontname = "Tests/fonts/10x20-ISO8859-1.pcf"

message = "hello, world"


pytestmark = skip_unless_feature("zlib")


def save_font(request: pytest.FixtureRequest, tmp_path: Path) -> str:
    with open(fontname, "rb") as test_file:
        font = PcfFontFile.PcfFontFile(test_file)
    assert isinstance(font, FontFile.FontFile)
    # check the number of characters in the font
    assert len([_f for _f in font.glyph if _f]) == 223

    tempname = str(tmp_path / "temp.pil")

    def delete_tempfile() -> None:
        try:
            os.remove(tempname[:-4] + ".pbm")
        except OSError:
            pass  # report?

    request.addfinalizer(delete_tempfile)
    font.save(tempname)

    with Image.open(tempname.replace(".pil", ".pbm")) as loaded:
        assert_image_equal_tofile(loaded, "Tests/fonts/10x20.pbm")

    with open(tempname, "rb") as f_loaded:
        with open("Tests/fonts/10x20.pil", "rb") as f_target:
            assert f_loaded.read() == f_target.read()
    return tempname


def test_sanity(request: pytest.FixtureRequest, tmp_path: Path) -> None:
    save_font(request, tmp_path)


def test_less_than_256_characters() -> None:
    with open("Tests/fonts/10x20-ISO8859-1-fewer-characters.pcf", "rb") as test_file:
        font = PcfFontFile.PcfFontFile(test_file)
    assert isinstance(font, FontFile.FontFile)
    # check the number of characters in the font
    assert len([_f for _f in font.glyph if _f]) == 127


def test_invalid_file() -> None:
    with open("Tests/images/flower.jpg", "rb") as fp:
        with pytest.raises(SyntaxError):
            PcfFontFile.PcfFontFile(fp)


def test_draw(request: pytest.FixtureRequest, tmp_path: Path) -> None:
    tempname = save_font(request, tmp_path)
    font = ImageFont.load(tempname)
    im = Image.new("L", (130, 30), "white")
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), message, "black", font=font)
    assert_image_similar_tofile(im, "Tests/images/test_draw_pbm_target.png", 0)


def test_textsize(request: pytest.FixtureRequest, tmp_path: Path) -> None:
    tempname = save_font(request, tmp_path)
    font = ImageFont.load(tempname)
    for i in range(255):
        (ox, oy, dx, dy) = font.getbbox(chr(i))
        assert ox == 0
        assert oy == 0
        assert dy == 20
        assert dx in (0, 10)
        assert font.getlength(chr(i)) == dx
    for i in range(len(message)):
        msg = message[: i + 1]
        assert font.getlength(msg) == len(msg) * 10
        assert font.getbbox(msg) == (0, 0, len(msg) * 10, 20)


def _test_high_characters(
    request: pytest.FixtureRequest, tmp_path: Path, message: AnyStr
) -> None:
    tempname = save_font(request, tmp_path)
    font = ImageFont.load(tempname)
    im = Image.new("L", (750, 30), "white")
    draw = ImageDraw.Draw(im)
    draw.text((0, 0), message, "black", font=font)
    assert_image_similar_tofile(im, "Tests/images/high_ascii_chars.png", 0)


def test_high_characters(request: pytest.FixtureRequest, tmp_path: Path) -> None:
    message = "".join(chr(i + 1) for i in range(140, 232))
    _test_high_characters(request, tmp_path, message)
    # accept bytes instances.
    _test_high_characters(request, tmp_path, message.encode("latin1"))



from __future__ import annotations

import warnings

import pytest

from PIL import Image, PsdImagePlugin

from .helper import assert_image_equal_tofile, assert_image_similar, hopper, is_pypy

test_file = "Tests/images/hopper.psd"


def test_sanity() -> None:
    with Image.open(test_file) as im:
        im.load()
        assert im.mode == "RGB"
        assert im.size == (128, 128)
        assert im.format == "PSD"
        assert im.get_format_mimetype() == "image/vnd.adobe.photoshop"

        im2 = hopper()
        assert_image_similar(im, im2, 4.8)


@pytest.mark.skipif(is_pypy(), reason="Requires CPython")
def test_unclosed_file() -> None:
    def open() -> None:
        im = Image.open(test_file)
        im.load()

    with pytest.warns(ResourceWarning):
        open()


def test_closed_file() -> None:
    with warnings.catch_warnings():
        warnings.simplefilter("error")

        im = Image.open(test_file)
        im.load()
        im.close()


def test_context_manager() -> None:
    with warnings.catch_warnings():
        warnings.simplefilter("error")

        with Image.open(test_file) as im:
            im.load()


def test_invalid_file() -> None:
    invalid_file = "Tests/images/flower.jpg"

    with pytest.raises(SyntaxError):
        PsdImagePlugin.PsdImageFile(invalid_file)


def test_n_frames() -> None:
    with Image.open("Tests/images/hopper_merged.psd") as im:
        assert im.n_frames == 1
        assert not im.is_animated

    for path in [test_file, "Tests/images/negative_layer_count.psd"]:
        with Image.open(path) as im:
            assert im.n_frames == 2
            assert im.is_animated


def test_eoferror() -> None:
    with Image.open(test_file) as im:
        # PSD seek index starts at 1 rather than 0
        n_frames = im.n_frames + 1

        # Test seeking past the last frame
        with pytest.raises(EOFError):
            im.seek(n_frames)
        assert im.tell() < n_frames

        # Test that seeking to the last frame does not raise an error
        im.seek(n_frames - 1)


def test_seek_tell() -> None:
    with Image.open(test_file) as im:
        layer_number = im.tell()
        assert layer_number == 1

        with pytest.raises(EOFError):
            im.seek(0)

        im.seek(1)
        layer_number = im.tell()
        assert layer_number == 1

        im.seek(2)
        layer_number = im.tell()
        assert layer_number == 2


def test_seek_eoferror() -> None:
    with Image.open(test_file) as im:
        with pytest.raises(EOFError):
            im.seek(-1)


def test_open_after_exclusive_load() -> None:
    with Image.open(test_file) as im:
        im.load()
        im.seek(im.tell() + 1)
        im.load()


def test_rgba() -> None:
    with Image.open("Tests/images/rgba.psd") as im:
        assert_image_equal_tofile(im, "Tests/images/imagedraw_square.png")


def test_negative_top_left_layer() -> None:
    with Image.open("Tests/images/negative_top_left_layer.psd") as im:
        assert im.layers[0][2] == (-50, -50, 50, 50)


def test_layer_skip() -> None:
    with Image.open("Tests/images/five_channels.psd") as im:
        assert im.n_frames == 1


def test_icc_profile() -> None:
    with Image.open(test_file) as im:
        assert "icc_profile" in im.info

        icc_profile = im.info["icc_profile"]
        assert len(icc_profile) == 3144


def test_no_icc_profile() -> None:
    with Image.open("Tests/images/hopper_merged.psd") as im:
        assert "icc_profile" not in im.info


def test_combined_larger_than_size() -> None:
    # The combined size of the individual parts is larger than the
    # declared 'size' of the extra data field, resulting in a backwards seek.

    # If we instead take the 'size' of the extra data field as the source of truth,
    # then the seek can't be negative
    with pytest.raises(OSError):
        with Image.open("Tests/images/combined_larger_than_size.psd"):
            pass


@pytest.mark.parametrize(
    "test_file,raises",
    [
        ("Tests/images/timeout-c8efc3fded6426986ba867a399791bae544f59bc.psd", OSError),
        ("Tests/images/timeout-dedc7a4ebd856d79b4359bbcc79e8ef231ce38f6.psd", OSError),
    ],
)
def test_crashes(test_file: str, raises: type[Exception]) -> None:
    with open(test_file, "rb") as f:
        with pytest.raises(raises):
            with Image.open(f):
                pass


@pytest.mark.parametrize(
    "test_file",
    [
        "Tests/images/timeout-1ee28a249896e05b83840ae8140622de8e648ba9.psd",
        "Tests/images/timeout-598843abc37fc080ec36a2699ebbd44f795d3a6f.psd",
    ],
)
def test_layer_crashes(test_file: str) -> None:
    with open(test_file, "rb") as f:
        with Image.open(f) as im:
            with pytest.raises(SyntaxError):
                im.layers



from __future__ import annotations

import pytest

from PIL.GimpPaletteFile import GimpPaletteFile


def test_sanity() -> None:
    with open("Tests/images/test.gpl", "rb") as fp:
        GimpPaletteFile(fp)

    with open("Tests/images/hopper.jpg", "rb") as fp:
        with pytest.raises(SyntaxError):
            GimpPaletteFile(fp)

    with open("Tests/images/bad_palette_file.gpl", "rb") as fp:
        with pytest.raises(SyntaxError):
            GimpPaletteFile(fp)

    with open("Tests/images/bad_palette_entry.gpl", "rb") as fp:
        with pytest.raises(ValueError):
            GimpPaletteFile(fp)


def test_get_palette() -> None:
    # Arrange
    with open("Tests/images/custom_gimp_palette.gpl", "rb") as fp:
        palette_file = GimpPaletteFile(fp)

    # Act
    palette, mode = palette_file.getpalette()

    # Assert
    assert mode == "RGB"



from __future__ import annotations

from PIL import Image, ImageDraw, ImageFont

font = ImageFont.truetype("Tests/fonts/NotoSans-Regular.ttf", 16)


def test(anchor: str) -> Image.Image:
    im = Image.new("RGBA", (200, 100), "white")
    d = ImageDraw.Draw(im)
    d.line(((100, 0), (100, 100)), "gray")
    d.line(((0, 50), (200, 50)), "gray")
    d.text((100, 50), f"{anchor} example", "black", font, anchor)
    return im


if __name__ == "__main__":
    im = Image.new("RGBA", (600, 300), "white")
    d = ImageDraw.Draw(im)
    for y, row in enumerate(
        (("ma", "mt", "mm"), ("ms", "mb", "md"), ("ls", "ms", "rs"))
    ):
        for x, anchor in enumerate(row):
            im.paste(test(anchor), (x * 200, y * 100))
            if x != 0:
                d.line(((x * 200, y * 100), (x * 200, (y + 1) * 100)), "black", 3)
            if y != 0:
                d.line(((x * 200, y * 100), ((x + 1) * 200, y * 100)), "black", 3)
    im.save("docs/example/anchors.webp")
    im.show()



from __future__ import annotations

import os
import sys
from io import BytesIO
from pathlib import Path

from PIL import Image, PSDraw


def _create_document(ps: PSDraw.PSDraw) -> None:
    title = "hopper"
    box = (1 * 72, 2 * 72, 7 * 72, 10 * 72)  # in points

    ps.begin_document(title)

    # draw diagonal lines in a cross
    ps.line((1 * 72, 2 * 72), (7 * 72, 10 * 72))
    ps.line((7 * 72, 2 * 72), (1 * 72, 10 * 72))

    # draw the image (75 dpi)
    with Image.open("Tests/images/hopper.ppm") as im:
        ps.image(box, im, 75)
    ps.rectangle(box)

    # draw title
    ps.setfont("Courier", 36)
    ps.text((3 * 72, 4 * 72), title)

    ps.end_document()


def test_draw_postscript(tmp_path: Path) -> None:
    # Based on Pillow tutorial, but there is no textsize:
    # https://pillow.readthedocs.io/en/latest/handbook/tutorial.html#drawing-postscript

    # Arrange
    tempfile = str(tmp_path / "temp.ps")
    with open(tempfile, "wb") as fp:
        # Act
        ps = PSDraw.PSDraw(fp)
        _create_document(ps)

    # Assert
    # Check non-zero file was created
    assert os.path.isfile(tempfile)
    assert os.path.getsize(tempfile) > 0


def test_stdout() -> None:
    # Temporarily redirect stdout
    old_stdout = sys.stdout

    class MyStdOut:
        buffer = BytesIO()

    mystdout = MyStdOut()

    sys.stdout = mystdout

    ps = PSDraw.PSDraw()
    _create_document(ps)

    # Reset stdout
    sys.stdout = old_stdout

    assert mystdout.buffer.getvalue() != b""
